/* vim: set sw=4 sts=4 et foldmethod=syntax : */

/*
 * Copyright (c) 2006 Ciaran McCreesh <ciaranm@ciaranm.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the library nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef LIBWRAPITER_GUARD_LIBWRAPITER_LIBWRAPITER_OUTPUT_ITERATOR_HH
#define LIBWRAPITER_GUARD_LIBWRAPITER_LIBWRAPITER_OUTPUT_ITERATOR_HH 1

#include <libwrapiter/libwrapiter_fwd.hh>
#include <iterator>

/** \file
 * Header for the OutputIterator class and its implementation.
 */

namespace libwrapiter
{
    /**
     * The underlying iterator wrapper for an OutputIterator.
     */
    template <typename Type_>
    class OutputIteratorBase
    {
        private:
            ///\name Disallow copying and assignment
            ///\{

            OutputIteratorBase(const OutputIteratorBase &);
            OutputIteratorBase operator= (const OutputIteratorBase &);

            ///\}

        protected:
            ///\name Basic operations
            ///\{

            OutputIteratorBase()
            {
            }

            ///\}

        public:
            ///\name Basic operations
            ///\{

            /**
             * Copy ourself.
             *
             * This can't be a regular copy constructor since copying removes
             * subclassing.
             */
            virtual OutputIteratorBase<Type_> * clone() = 0;

            /**
             * Destructor.
             */
            virtual ~OutputIteratorBase()
            {
            }

            /**
             * Assign, used by the operator= method.
             */
            virtual void assign(const Type_ & value) = 0;

            ///\}

            ///\name Increment operators
            ///\{

            /**
             * Increment, used by both operator++ methods.
             */
            virtual void increment() = 0;

            ///\}
    };

    /**
     * The implementation of an OutputIteratorBase for a particular iterator
     * type.
     */
    template <typename Type_, typename Iter_>
    class OutputIteratorBaseImpl :
        public OutputIteratorBase<Type_>
    {
        private:
            /// Our underlying iterator.
            Iter_ _iter;

        public:
            ///\name Basic operations
            ///\{

            /**
             * Constructor, from an iterator.
             */
            OutputIteratorBaseImpl(const Iter_ & i) :
                _iter(i)
            {
            }

            ///\}

            virtual OutputIteratorBase<Type_> * clone()
            {
                return new OutputIteratorBaseImpl<Type_, Iter_>(_iter);
            }

            virtual void assign(const Type_ & value)
            {
                *_iter = value;
            }

            virtual void increment()
            {
                ++_iter;
            }
    };

    /**
     * A generic insert iterator that inserts an item of type Type_.
     *
     * The template is parameterised by OurType_, which should be a unique
     * class (usually the owning class) for each iterator. This is to
     * prevent comparing, say, an OutputIterator<int> implemented in terms of a
     * set iterator with an OutputIterator<int> implemented in terms of a list
     * iterator. The Type_ parameter is the type of the underlying item.
     */
    template <typename OurType_, typename Type_>
    class OutputIterator :
        public std::iterator<std::output_iterator_tag, Type_>
    {
        private:
            /// Our underlying handler.
            OutputIteratorBase<Type_> * _base;

        public:
            ///\name STL related typedefs
            ///\{

            /**
             * Fake a container_type, which is required by various STL things.
             */
            struct container_type
            {
                ///\name STL related typedefs
                ///\{

                typedef Type_ value_type;

                ///\}
            };

            ///\}

            ///\name Basic operations
            ///\{

            /**
             * Constructor, from an iterator.
             */
            template <typename Iter_>
            explicit OutputIterator(const Iter_ & iter) :
                _base(new OutputIteratorBaseImpl<Type_, Iter_>(iter))
            {
            }

            /**
             * Copy constructor.
             */
            OutputIterator(const OutputIterator<OurType_, Type_> & other) :
                _base(other._base->clone())
            {
            }

            /**
             * Assignment operator.
             */
            const OutputIterator & operator= (const Type_ & value)
            {
                _base->assign(value);
                return *this;
            }

            /**
             * Destructor.
             */
            ~OutputIterator()
            {
                delete _base;
            }

            ///\}

            ///\name Dereference operators
            ///\{

            /**
             * Dereference operator.
             */
            OutputIterator & operator* ()
            {
                return *this;
            }

            /**
             * Dereference operator.
             */
            const OutputIterator * operator-> () const
            {
                return this;
            }

            /**
             * Dereference operator.
             */
            const OutputIterator & operator* () const
            {
                return *this;
            }

            /**
             * Dereference operator.
             */
            OutputIterator * operator-> ()
            {
                return this;
            }

            ///\}

            ///\name Increment operators
            ///\{

            /**
             * Preincrement.
             */
            OutputIterator<OurType_, Type_> & operator++ ()
            {
                _base->increment();
                return *this;
            }

            /**
             * Postincrement.
             */
            OutputIterator<OurType_, Type_> operator++ (int)
            {
                OutputIterator<OurType_, Type_> result(*this);
                _base->increment();
                return result;
            }

            ///\}
    };
}

#endif
